import unittest
from uuid import UUID

from mock import patch
from fastapi.testclient import TestClient

from delta.run.api import app, RunBase, RunBaseCreate
from delta.run.orchestrator import DeltaOrchestratorService

client = TestClient(app)


class TestApi(unittest.TestCase):

    def fake_get_run_by_id(self, run_id: str):
        return {"run": '"test_ok"', "inputs": 'in', "outputs": 'out'}

    def fake_get_runs(self):
        return {"runs": [{"run": '"test_ok"',
                          "inputs": 'in',
                          "outputs": 'out'},
                         {"run": '"test_ok"',
                          "inputs": 'in',
                          "outputs": 'out'}]}

    def fake_get_runs_ko(self):
        raise Exception("Error when get runs")

    async def fake_orchestrate(self, owner, twin_id, user_inputs):
        return {"run": '"test_ok"', "inputs": 'in', "outputs": "out"}

    async def fake_orchestrate_KO(self, owner, twin_id, user_inputs):
        raise Exception("Error when create run")

    async def fake_actifact(self):
        return {"run": '"test_ok"', "inputs": 'in', "outputs": 'out'}

    @patch.object(DeltaOrchestratorService, 'get_run_by_id',
                  fake_get_run_by_id)
    def test_get_runs(self):
        response = client.get("/runs/61cdd9a3-30e5-449c-ad23-e83d24fb25c9")
        print(response)
        print(response.json())
        self.assertIn("run", response.json())
        self.assertNotIn("test_ko", response.json())
        pass

    @patch.object(DeltaOrchestratorService, 'get_run_by_id',
                  fake_get_run_by_id)
    def test_get_404(self):
        response = client.get("/runs_twin")
        print(response)
        print(response.json())
        self.assertEqual(response.status_code, 404)
        pass

    @patch.object(DeltaOrchestratorService, 'get_run_by_id',
                  fake_get_run_by_id)
    def test_get_runs_error(self):
        response = client.get("/runs/11345")
        print(response)
        print(response.json())
        self.assertEqual(response.status_code, 422)
        pass

    @patch.object(DeltaOrchestratorService, 'get_runs', fake_get_runs)
    def test_get_runs_all(self):
        response = client.get("/runs")
        print(response)
        print(response.json())
        self.assertEqual(response.status_code, 200)
        pass

    @patch.object(DeltaOrchestratorService, 'get_runs', fake_get_runs_ko)
    def test_get_runs_all_exception(self):

        response = client.get("/runs")
        print(response)
        print(response.json())
        self.assertEqual(response.status_code, 500)

    @patch.object(DeltaOrchestratorService, 'orchestrate', fake_orchestrate)
    def test_create(self):

        run_context = RunBaseCreate(
            owner='test',
            twin_id='61cdd9a3-30e5-449c-ad23-e83d24fb25c9',
            user_inputs={})
        print(run_context.model_dump_json())
        response = client.post("/runs", data=run_context.model_dump_json())
        print(response)
        print(response.json())
        self.assertEqual(response.status_code, 201)

    @patch.object(DeltaOrchestratorService, 'orchestrate', fake_orchestrate_KO)
    def test_create_exception(self):

        run_context = RunBaseCreate(
            owner='test',
            twin_id='61cdd9a3-30e5-449c-ad23-e83d24fb25c9',
            user_inputs={})
        print(run_context.model_dump_json())
        response = client.post("/runs", data=run_context.model_dump_json())
        print(response)
        print(response.json())
        self.assertEqual(response.status_code, 500)
        pass

    @patch.object(DeltaOrchestratorService, 'orchestrate', fake_orchestrate)
    def test_create_ko(self):

        run_context = RunBase(
            owner='test',
            twin_id=UUID('61cdd9a3-30e5-449c-ad23-e83d24fb25c9'))
        print(run_context.model_dump_json())
        response = client.post("/runs", data=run_context.model_dump_json())
        print(response)
        print(response.json())
        self.assertEqual(response.status_code, 422)

    @patch.object(DeltaOrchestratorService, 'orchestrate', fake_actifact)
    def test_create_artifact(self):

        response = (
            client.post(
                "/runs/61cdd9a3-30e5-449c-ad23-e83d24fb25c9/artifacts"))
        print(response)
        print(response.json())
        self.assertEqual(response.status_code, 201)

    @patch.object(DeltaOrchestratorService, 'orchestrate', fake_orchestrate)
    def test_create_artifact_ko(self):

        response = client.post(
            "/runs/61cdd9a3-30e5-449c-ad23-e8c9/artifacts")
        print(response)
        print(response.json())
        self.assertEqual(response.status_code, 422)

    @patch.object(DeltaOrchestratorService, 'orchestrate', fake_actifact)
    def test_get_artifact(self):

        response = (
            client.get(
                "/runs/61cdd9a3-30e5-449c-ad23-e83d24fb25c9/artifacts"))
        print(response)
        print(response.json())
        self.assertEqual(response.status_code, 200)



import abc
import os
from typing import Any, Union

from delta.manifest.parser import InputModel
from delta.run.api.model import DataParameterModel


class Placeholder(abc.ABC):
    @abc.abstractmethod
    def evaluate(self) -> Any:
        raise NotImplementedError


class InputModelPlaceholder(Placeholder):
    _separator: str = " "

    def __init__(self, input_model: InputModel, value: Any):
        super().__init__()
        self._input = input_model
        self._value = value
        self._factory = {
            "boolean": self.__evaluate_boolean,
            "integer": self.__evaluate_number,
            "number": self.__evaluate_number,
            "string": self.__evaluate_string,
            "Data": self.__evaluate_data,
            "stdout": self.__evaluate_data,
        }

    def __evaluate_boolean(self) -> str:
        value = self._value or self._input.value
        if value and self._input.prefix:
            return self._input.prefix
        return ""

    def __evaluate_number(self) -> str:
        value = self._value or self._input.value
        if value:
            if self._input.prefix:
                return f"{self._input.prefix}{self._separator}{value}"
            return str(value.value)
        return ""

    def __evaluate_string(self) -> str:
        value = self._value or self._input.value
        if value:
            if self._input.prefix:
                return f"{self._input.prefix}{self._separator}{value}"
            return value.value
        return ""

    def __evaluate_data(self) -> str:
        value = self._value or self._input.value
        if value:
            if self._input.prefix:
                return f"{self._input.prefix}{self._separator}{value.path}"
            return os.path.join("/s3", value.path)
        return ""

    def evaluate(self) -> str:
        try:
            return self._factory[self._input.type]()
        except KeyError:
            return ""


class JobBuilder(abc.ABC):
    @staticmethod
    def placeholder_factory(model_part: InputModel, value: Any) -> Placeholder:
        if isinstance(model_part, InputModel):
            return InputModelPlaceholder(model_part, value)
        raise NotImplementedError

    @abc.abstractmethod
    def generate_dockerfile(self) -> str:
        raise NotImplementedError

    @staticmethod
    def generate_dockerfile_ctx() -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def build_command(self):
        raise NotImplementedError



import asyncio
import collections

import ansible_runner
import kubernetes

from delta.run.config import Settings
from delta.run.job import Job, JobService, RunStatus


K8sContainerState = collections.namedtuple("K8sState", ["state", "reason"])


class DeltaJobService(JobService):
    def __init__(self):
        super().__init__()
        self._config = Settings()
        # kubernetes client
        k8s_conf = self.__prepare_kubernetes_config()
        kubernetes.config.load_kube_config_from_dict(k8s_conf)
        self._k8s_cli = kubernetes.client.CoreV1Api()
        self._env = self.__prepare_ansible_config()

    def __prepare_ansible_config(self) -> dict:
        return {
            "k8s": {
                "namespace": self._config.k8s_namespace,
                "context": self._config.k8s_context,
                "cluster": {
                    "name": self._config.k8s_cluster_name,
                    "cert_auth": self._config.k8s_cluster_cert_auth,
                    "server": self._config.k8s_cluster_server,
                },
                "user": {
                    "name": self._config.k8s_user_name,
                    "client_cert_data": self._config.k8s_user_cli_cert,
                    "client_key_data": self._config.k8s_user_cli_key,
                },
            },
            "storage": {
                "data_path": "/s3",
                "s3": {
                    "bucket": {"name": self._config.s3_bucket},
                    "endpoint": self._config.s3_endpoint,
                    "region": self._config.s3_region,
                    "accesskeyid": self._config.s3_access_key,
                    "secretaccesskey": self._config.s3_secret_access_key,
                },
            },
        }

    def __prepare_kubernetes_config(self) -> dict:
        """
        Prepares kubernetes config
        """
        return {
            "apiVersion": "v1",
            "kind": "Config",
            "clusters": [
                {
                    "name": self._config.k8s_cluster_name,
                    "cluster": {
                        "certificate-authority-data":
                            self._config.k8s_cluster_cert_auth,
                        "server": self._config.k8s_cluster_server,
                    },
                }
            ],
            "contexts": [
                {
                    "name": self._config.k8s_context,
                    "context": {
                        "user": self._config.k8s_user_name,
                        "cluster": self._config.k8s_cluster_name,
                    },
                }
            ],
            "current-context": self._config.k8s_context,
            "preferences": {},
            "users": [
                {
                    "name": self._config.k8s_user_name,
                    "user": {
                        "client-certificate-data":
                            self._config.k8s_user_cli_cert,
                        "client-key-data": self._config.k8s_user_cli_key,
                    },
                }
            ],
        }

    @staticmethod
    def __retrieve_container_state_from_pod(
        pod: kubernetes.client.models.V1Pod,
    ) -> K8sContainerState:
        k8s_state = pod.status.container_statuses[0].state
        if k8s_state.running is not None:
            return K8sContainerState("running", None)
        elif k8s_state.terminated is not None:
            return K8sContainerState("terminated", k8s_state.terminated.reason)
        elif k8s_state.waiting is not None:
            return K8sContainerState("waiting", k8s_state.waiting.reason)
        raise ValueError(f"Unknown container state from pod: {pod}")

    @staticmethod
    def __check_container_state(state: K8sContainerState) -> int | None:
        match state:
            case (
                K8sContainerState("running", None)
                | K8sContainerState("terminated", "Terminating")
                | K8sContainerState("waiting", "ContainerCreating")
                | K8sContainerState("waiting", "PodInitializing")
            ):
                return None
            case (
                K8sContainerState("waiting", "ImagePullBackOff")
                | K8sContainerState("waiting", "ErrImagePull")
                | K8sContainerState("waiting", "CrashLoopBackOff")
                | K8sContainerState("waiting", "ImageInspectError")
                | K8sContainerState("waiting", "CreateContainerConfigError")
                | K8sContainerState("waiting", "ErrImageNeverPull")
                | K8sContainerState("waiting", "InvalidImageName")
                | K8sContainerState("terminated", "Error")
                | K8sContainerState("terminated", "OOMKilled")
                | K8sContainerState("terminated", "ContainerCannotRun")
                | K8sContainerState("terminated", "DeadlineExceeded")
                | K8sContainerState("terminated", "ImageGCFailed")
                | K8sContainerState("terminated", "ContainerStatusUnknown")
            ):
                return 1
            case K8sContainerState("terminated", "Completed"):
                return 0
            case _:
                raise ValueError(f"Unknown container state: {state}")

    async def execute_job(self, job: Job, **kwargs):
        # prepare environment variables to execute ansible playbook run_job
        inputs = ";".join(map(lambda x: f"{x.src},{x.dest}", job.inputs))
        outputs = ";".join(map(lambda x: f"{x.src},{x.dest}", job.outputs))
        self._env["job"] = {
            "name": str(job.id),
            "image": f"{self._config.image_repo_hostname}/{job.image.tag}",
            "cmd": job.command,
            "working_directory": job.working_dir,
            "resources": {
                "inputs": inputs,
                "outputs": outputs,
            },
            "registry": {
                "url": self._config.image_repo_hostname,
                "login": self._config.image_repo_username,
                "password": self._config.image_repo_password,
            },
        }

        # executing run_job playbook from gael.delta
        try:
            r = await asyncio.to_thread(
                ansible_runner.run,
                private_data_dir="gael.delta",
                extravars=self._env,
                playbook="playbooks/run_job.yml",
                suppress_env_files=True,
                quiet=True,
            )
            if r.rc != 0:
                self._logger.error(
                    "Ansible failure: %d\n stdout:\n%s\n stderr:\n%s",
                    r.rc, r.stdout, r.stderr
                )
                raise RuntimeError(f"Failed to start job via ansible: {r.rc}")
            self._jobs[job.id] = job
        except Exception as e:
            self._logger.info(f"Job({job.id}) failed: {e}")
            self._logger.error(e)
            job.status = RunStatus.ERROR

    def check_jobs(self):
        updated = []
        for key, job in self._jobs.items():
            try:
                pod = self._k8s_cli.read_namespaced_pod(
                    str(key), self._config.k8s_namespace)
                state = self.__retrieve_container_state_from_pod(pod)
                status = self.__check_container_state(state)
                if status is not None:
                    updated.append(key)
                    if status == 0:
                        job.status = RunStatus.SUCCESS
                    elif status == 1:
                        job.status = RunStatus.ERROR
            except kubernetes.client.exceptions.ApiException as e:
                self._logger.error(
                    f"Failed to check job status: {key}, reason: {e.reason}"
                )
            except ValueError as e:
                self._logger.error(f"Unknown job status: {e}")
            except Exception:
                self._logger.error("Pod initializing...")
        # unwatch updated jobs
        for key in updated:
            del self._jobs[key]
            self._k8s_cli.delete_namespaced_pod(
                str(key),
                self._config.k8s_namespace
            )

    def shutdown(self):
        pass



from typing import Annotated

from pydantic import Field, HttpUrl, PostgresDsn, UrlConstraints
from pydantic_core import MultiHostUrl
from pydantic_settings import BaseSettings, SettingsConfigDict


SQLiteUrl = Annotated[
    MultiHostUrl,
    UrlConstraints(
        host_required=False,
        allowed_schemes=['sqlite', 'sqlite3'],
    )
]


class Settings(BaseSettings):
    model_config = SettingsConfigDict(
        case_sensitive=False,
        env_prefix='DELTA_',
    )

    # Deltatwin Run
    api_url: HttpUrl
    run_limit: int = Field(default=10)
    eviction_active: bool = Field(default=False)
    eviction_keep_period: int = Field(default=48)
    database_url: SQLiteUrl | PostgresDsn = Field(
        default="sqlite:///.delta-run.db?check_same_thread=false"
    )
    database_show_sql: bool = Field(default=False)
    page_limit: int = Field(default=100)

    # DeltaTwin container image registry
    image_repo_hostname: str
    image_repo_username: str
    image_repo_password: str

    # DeltaTwin Run output storage
    s3_endpoint: str
    s3_region: str
    s3_access_key: str
    s3_secret_access_key: str
    s3_bucket: str

    # kubernetes executor
    k8s_context: str
    k8s_namespace: str
    k8s_cluster_name: str
    k8s_cluster_cert_auth: str
    k8s_cluster_server: str
    k8s_user_name: str
    k8s_user_cli_cert: str
    k8s_user_cli_key: str



import asyncio
import logging
import os
import re
import tempfile
import uuid
import requests as req
from datetime import datetime, timedelta
from typing import IO, List

from sqlalchemy.exc import NoResultFound

import delta.run.db.daos as dao
from delta.run.api.model import (
    RunContextInsertModel,
    RunContextModel,
    RunContextUpdateModel,
    model_param_to_orm_param,
)
from delta.run.config import Settings
from delta.run.db.database import Session
from delta.run.db.orm import (
    DataParameter,
    ParameterKind,
    RunContext,
    RunStatus,
)
from delta.run.job.delta_service import DeltaJobService
from delta.run.orchestrator import DeltaOrchestrator
from delta.run.storage_manager import DeltaStorageManager
from delta.run.utils import DependencyResolver

logger = logging.getLogger("RunService")
logger.setLevel(logging.INFO)


class DeltaRunService:
    def __init__(self, storage: DeltaStorageManager):
        self._storage = storage
        self._dependency_resolver = DependencyResolver()
        self._run = set()

    @staticmethod
    def __download_deltatwin_file(
            token: str,
            deltatwin_id: str,
            deltatwin_version: str,
            file: str,
            destination_dir: str) -> None:
        url = (
            f"{Settings().api_url}deltatwins/"
            f"{deltatwin_id}/files/{file}?version={deltatwin_version}"
        )
        response = req.get(
            url.format('manifest'),
            # token format: 'Bearer <OAuth2_token>'
            headers={'Authorization': token},
            stream=True
        )
        if response.status_code != 200:
            logger.error(
                "Failed to retrieve file `%s` from deltatwin: "
                "%s v%s (%s, %s, %d)",
                file, deltatwin_id, deltatwin_version, response.url,
                response.request.method, response.status_code
            )
            raise RuntimeError(
                f"Enable to retrieve {file}  file from deltatwin: "
                f"{deltatwin_id}({deltatwin_version})"
            )

        filename = re.findall(
            'filename="?([^"]+)', response.headers["Content-Disposition"]
        )[0]
        destination = os.path.join(destination_dir, filename)
        with open(destination, "wb") as f:
            for chunk in response.iter_content(chunk_size=4096):
                f.write(chunk)

    def __download_twin(self, token, twin_id, twin_version):
        # Downloading manifest and workflow in a temporary directory
        with tempfile.TemporaryDirectory() as tmp_dir:
            self.__download_deltatwin_file(
                token, twin_id, twin_version, 'manifest', tmp_dir
            )
            self.__download_deltatwin_file(
                token, twin_id, twin_version, 'workflow', tmp_dir
            )
            self._storage.push_deltatwin(tmp_dir, twin_id, twin_version)

    def __download_twin_with_dependencies(
            self, token, twin_id, twin_version, visited=None
    ):
        if visited is None:
            visited = set()

        if twin_id in visited:
            return

        visited.add((twin_id, twin_version))
        if not self._storage.has_deltatwin(twin_id, twin_version):
            self.__download_twin(token, twin_id, twin_version)

        manifest = self._storage.get_deltatwin_manifest(twin_id, twin_version)
        if manifest.dependencies is not None:
            for dependency in manifest.dependencies.values():
                self._dependency_resolver.add_dependency(
                    twin_id, (dependency.id, dependency.version))
                if self._dependency_resolver.check_dependency_cycles(twin_id):
                    raise RuntimeError("Dependency cycle detected")
                self.__download_twin_with_dependencies(
                    token, dependency.id, dependency.version, visited)

    @staticmethod
    def __on_update_run_context(run_id: str, context: RunContextUpdateModel):
        with Session().begin() as tx:
            dao.update_run(tx.session, run_id, context)

    @staticmethod
    def __on_create_sub_run_context(run_id: str, context: RunContext):
        logger.info('Prepare sub run %s from %s', run_id, context.id)
        with Session().begin() as tx:
            dao.add_run(tx.session, context)

    async def orchestrate(self, run_ctx: RunContextModel):
        async with DeltaJobService() as executor:
            try:
                manifest = self._storage.get_deltatwin_manifest(
                    run_ctx.deltatwin_id, run_ctx.deltatwin_version
                )
                workflow = self._storage.get_deltatwin_workflow(
                    run_ctx.deltatwin_id, run_ctx.deltatwin_version
                )
                orch = DeltaOrchestrator(
                    manifest=manifest,
                    workflow=workflow,
                    run_ctx=run_ctx,
                    storage_manager=self._storage,
                    job_service=executor,
                    on_update_context=self.__on_update_run_context,
                    on_create_sub_context=self.__on_create_sub_run_context
                )
                await orch.run()
            except (ValueError, KeyError) as ex:
                run_ctx.status = RunStatus.ERROR
                if isinstance(ex, ValueError):
                    run_ctx.return_code = 311
                    run_ctx.message = "Invalid manifest content"
                if isinstance(ex, KeyError):
                    run_ctx.return_code = 310
                    run_ctx.message = (
                        f"Node {ex.args[0]} not found in manifest"
                    )
                with Session().begin() as tx:
                    rcup = RunContextUpdateModel(
                        status=run_ctx.status,
                        return_code=run_ctx.return_code,
                        message=run_ctx.message,
                    )
                    dao.update_run(tx.session, run_ctx.id, rcup)
                return
            finally:
                executor.shutdown()

    def add_run(self, token, ctx: RunContextInsertModel) -> RunContextModel:
        run_ctx = RunContext(
            id=str(uuid.uuid4()),
            deltatwin_id=ctx.deltatwin_id,
            deltatwin_version=ctx.deltatwin_version,
            owner=ctx.owner,
            status=RunStatus.CREATED,
            inputs=[
                model_param_to_orm_param(p, ParameterKind.INPUT)
                for p in ctx.inputs
            ],
            date_created=ctx.date_created,
        )
        # download and push on the storage the deltatwin if necessary
        self.__download_twin_with_dependencies(
            token, run_ctx.deltatwin_id, run_ctx.deltatwin_version
        )

        with Session().begin() as tx:
            run = RunContextModel.from_run_context(
                dao.add_run(tx.session, run_ctx)
            )
            # keep background task and remove it when finish
            task = asyncio.create_task(self.orchestrate(run))
            self._run.add(task)
            task.add_done_callback(self._run.discard)

            return run

    @staticmethod
    def get_runs(
        deltatwin_id: str = None,
        owner: str = None,
        status: RunStatus = None,
        limit: int = 0,
        offset: int = 0,
    ) -> List[RunContextModel]:
        with Session() as session:
            runs = dao.get_runs(
                session, deltatwin_id, owner, status, limit, offset
            )
            return [RunContextModel.from_run_context(r) for r in runs]

    @staticmethod
    def get_run_by_id(run_id: str) -> RunContextModel:
        with Session() as session:
            try:
                run_ctx = dao.get_run_by_id(session, run_id)
                return RunContextModel.from_run_context(run_ctx)
            except NoResultFound:
                raise ValueError(f"RunContext not found: {run_id}")

    async def get_run_output(self, run_id: str, output_name: str) -> IO:
        with Session() as session:
            run_ctx = dao.get_run_by_id(session, run_id)
            output = next(
                filter(lambda n: n.name == output_name, run_ctx.outputs), None
            )

            if output is None:
                raise RuntimeError(f"Output not found: {output_name}")

            if not isinstance(output, DataParameter):
                raise RuntimeError("Only output type Data can be requested")

            split = output.path.split("/", maxsplit=3)
            data = asyncio.to_thread(
                self._storage.get_data,
                run_id=run_ctx.id,
                job_id=split[2],
                basename=split[3],
            )
            return await data

    def remove_run(self, run_id: uuid.UUID):
        with Session() as session:
            try:
                dao.delete_run(session, str(run_id))
                self._storage.remove_run(run_id)
            except NoResultFound:
                raise ValueError(f"RunContext not found: {run_id}")

    def remove_old_runs(self, threshold_hours: int):
        now = datetime.utcnow()
        threshold_date = now - timedelta(hours=threshold_hours)
        with Session() as session:
            try:
                runs_to_remove = dao.get_runs_by_date(session, threshold_date)
                for run in runs_to_remove:
                    dao.delete_run(session, run.id)
                    self._storage.remove_run(uuid.UUID(run.id))

                logger.info(f"Successfully removed {len(runs_to_remove)} "
                            f"runs older than {threshold_date}.")
            except Exception as e:
                logger.error(f"Error occurred while removing runs: {str(e)}")

    def get_number_of_run_by_user(self, user):
        with Session() as session:
            try:
                count = dao.get_number_of_run_by_user(session, user)
                return count
            except Exception as e:
                logger.error(e)

    def set_running_runs_to_error_at_startup(self):
        with Session() as session:
            dao.set_running_runs_to_error_at_startup(session)



import json
import logging
import logging.config
import logging.handlers
import os
import pathlib
import uuid
from contextlib import asynccontextmanager
from typing import Annotated, List, Optional

import alembic.command
import alembic.config
import uvicorn
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from fastapi import FastAPI, Header, HTTPException, status
from pytz import utc
from starlette.responses import StreamingResponse

from delta.run.api.model import RunContextInsertModel, RunContextModel
from delta.run.config import Settings
from delta.run.db.orm import RunStatus
from delta.run.service import DeltaRunService
from delta.run.storage_manager import S3StorageManager


def setup_logging():
    config_file = pathlib.Path("log_config.json")
    with open(config_file) as f_in:
        config = json.load(f_in)
    logging.config.dictConfig(config)


scheduler_logger = logging.getLogger("apscheduler")
scheduler_logger.setLevel(logging.INFO)

LIMIT_DEFAULT = 100

settings = Settings()

alembic.command.upgrade(alembic.config.Config("alembic.ini"), "head")
storage = S3StorageManager(
    entry_point=settings.s3_endpoint,
    key=settings.s3_access_key,
    secret=settings.s3_secret_access_key,
    bucket=settings.s3_bucket
)
run_svc = DeltaRunService(storage)


def eviction():
    try:
        run_svc.remove_old_runs(threshold_hours=settings.eviction_keep_period)
    except Exception as e:
        scheduler_logger.error(f"Error occurred during eviction: {e}")


@asynccontextmanager
async def lifespan(app: FastAPI):
    # Startup logic
    scheduler = None
    if settings.eviction_active:
        scheduler = AsyncIOScheduler(timezone=utc)
        scheduler.add_job(
            eviction,
            'cron',
            day_of_week='*',
            hour=0,
            minute=0
        )
        scheduler.start()

    run_svc.set_running_runs_to_error_at_startup()

    try:
        yield
    finally:
        # Shutdown logic
        if scheduler and scheduler.running:
            scheduler.shutdown()


app = FastAPI(lifespan=lifespan)


def check_run_id(run_id: uuid.UUID):
    if not isinstance(run_id, uuid.UUID):
        raise HTTPException(
            status_code=422, detail="'run_id' must be a valid UUID"
        )


@app.post("/runs", status_code=status.HTTP_201_CREATED)
async def create_run(
    token: Annotated[Optional[str], Header()], run_model: RunContextInsertModel
) -> RunContextModel:
    try:
        if run_svc.get_number_of_run_by_user(run_model.owner) >= \
                settings.run_limit:
            raise HTTPException(
                status_code=429,
                detail=(
                    "You have reached the number of runs allowed for the "
                    "beta version, please try again later"
                ),
            )
        return run_svc.add_run(token, run_model)
    except RuntimeError as e:
        raise HTTPException(status_code=500, detail=e)


@app.get("/runs")
def read_runs(
    deltatwin_id: str = None, owner: str = None,
        status: RunStatus = None,
        limit: int = LIMIT_DEFAULT,
        offset: int = 0
) -> List[RunContextModel]:
    return run_svc.get_runs(deltatwin_id, owner, status, limit, offset)


@app.get("/runs/{run_id}")
def read_run(run_id: uuid.UUID) -> RunContextModel:
    check_run_id(run_id)
    try:
        rc = run_svc.get_run_by_id(str(run_id))
        return rc
    except ValueError:
        raise HTTPException(status_code=404, detail=f"Run {run_id} not found")


@app.delete("/runs/{run_id}", status_code=status.HTTP_204_NO_CONTENT)
def delete_run(run_id: uuid.UUID):
    try:
        run_svc.remove_run(run_id)
    except ValueError:
        raise HTTPException(status_code=404)


@app.get("/runs/{run_id}/output/{output_name}/info")
def get_output(run_id: str, output_name: str):
    run = run_svc.get_run_by_id(run_id)
    output = next(filter(lambda p: p.name == output_name, run.outputs))
    if output is not None:
        return output
    raise HTTPException(
        status_code=404, detail=f"No input found: {output_name}"
    )


@app.get("/runs/{run_id}/output/{output_name}/download")
async def download_output(run_id: str, output_name: str):
    run = run_svc.get_run_by_id(run_id)
    output = next(filter(lambda p: p.name == output_name, run.outputs))
    if output is None:
        raise HTTPException(status_code=404, detail="Output not found")

    basename = os.path.basename(output.path)
    stream_io = await run_svc.get_run_output(run_id, output_name)
    info = stream_io.info()
    response = StreamingResponse(
        stream_io,
        status_code=status.HTTP_200_OK,
        headers={
            "Content-Disposition": f"attachment; filename={basename}",
            "Content-Length": info['content-length'],
            "Content-Type": info['content-type'],
        },
        media_type=info['content-type'] or "application/octet-stream",
    )
    return response


if __name__ == "__main__":
    setup_logging()
    uvicorn.run(
        app,
        host="0.0.0.0",
        port=8000,
    )



class DependencyResolver:
    def __init__(self):
        self.dependencies = {}
        self.visited = set()

    def add_dependency(self, component, dependency):
        self.dependencies.setdefault(component, []).append(dependency)

    def check_dependency_cycles(self, node):
        try:
            self.visited.clear()
            self.check_cycle(node, set())
            return False
        except ValueError:
            return True

    def check_cycle(self, node, visited):

        if node in visited:
            raise ValueError("Dependency cycle detected, Unable to continue.")

        visited.add(node)

        dependencies = self.dependencies.get(node, [])
        for neighbor in dependencies:
            self.check_cycle(neighbor, visited.copy())
